# Molar Mass Calculator
# (c) 2010-2011, Denis G. Samsonenko <d.g.samsonenko@gmail.com>
# LGPL

# Main package

package MMCalc;

use 5.008009;

use strict;
no warnings;

use Glib qw/TRUE FALSE/;

use File::BaseDir qw(data_home data_files);
use File::Spec::Functions;
use Cwd qw(cwd realpath);
use FindBin qw($RealBin);
use File::Path;

use POSIX;
setlocale(LC_NUMERIC, 'C');

our @EXPORT = ();
our @EXPORT_OK = qw(init get_acronyms set_acronyms check_acronym write_acronyms get_version);
our %EXPORT_TAGS = (all => [ @EXPORT_OK ]);
use base qw(Exporter);

# Version
our $VERSION = '20170901';
sub get_version
{
  my $version = $VERSION;
  $version =~ s/^(\d{4})(\d{2})(\d{2})$/$1-$2-$3/;
  return $version;
}

# Global variables:

# array of acronyms: $acronyms[$i][0] = acronym, [1] = formula, [2] = name or comments
our @acronyms = ();

# list of elements: $list[$i][0] = atom number, [1] = atom label,
#                            [2] = atomic mass, [3] = mass error
our @list = ();

# get script dir
FindBin->again;
our $script_dir = $RealBin;

# get prefix dir
our $prefix = realpath(catfile($script_dir, '..'));

# get current dir
our $current_dir = cwd();

our $atoms_file = "";
our $acronyms_file = "";

# Constructor
sub new
{
  if ($#list < 0)
  {
    return undef unless(eval{init()});
  }
  my $class = shift;
  my $self = {
              formula => "",
              mass => 0,
              error => 0,
              electrons => 0,
              atoms => [],
              content => ""
             };
  bless $self, $class;
  return $self;
}


# Reset
sub reset
{
  my $self = shift;
  $self->{formula} = "";
  $self->{mass} = 0;
  $self->{error} = 0;
  $self->{electrons} = 0;
  $self->{atoms} = [];
  $self->{content} = "";
}


# Get methods:

sub get_mass
{
  my $self = shift;
  return $self->{mass};
}

sub get_error
{
  my $self = shift;
  return $self->{error};
}

sub get_electrons
{
  my $self = shift;
  return $self->{electrons};
}

sub get_atoms
{
  my $self = shift;
  return $self->{atoms};
}

sub get_formula
{
  my $self = shift;
  if ($self->{content})
  {
    return $self->{content};
  }
  else
  {
    my $atoms = $#{$self->{atoms}};
    return undef if (!defined($self) or ($atoms < 0));

    my $contlist = "";
    for (my $i = 0; $i <= $atoms; ++$i)
    {
      next unless ($self->{atoms}->[$i][1]);
      $contlist .= $self->{atoms}->[$i][0];
      $contlist .= $self->{atoms}->[$i][1] if ($self->{atoms}->[$i][1] != 1);
    }
    return $contlist;
  }
}


# Calculations:

# reading number
sub number
{
  my $char = shift;
  my $params = shift;
  $params->{nmode} = TRUE;
  $params->{numbers} .= $char;
}

# end of atom symbol
sub endatom
# returns TRUE if OK or FALSE if error
{
  my $params = shift;
  for (my $i = 0; $i <= $params->{uniqatoms}; ++$i )
  {
    ($params->{mode} = 3, return TRUE)
      if ($params->{atom}->[$params->{atoms}][0] eq $params->{uniqatom}->[$i][0]);
  }
  for (my $i = 0; $i <= $#list; ++$i)
  {
    if ($params->{atom}->[$params->{atoms}][0] eq $list[$i][1])
    {
      ++$params->{uniqatoms};
      $params->{uniqatom}->[$params->{uniqatoms}][0] = $params->{atom}->[$params->{atoms}][0];
      $params->{uniqatom}->[$params->{uniqatoms}][1] = 0;
      $params->{uniqatom}->[$params->{uniqatoms}][2] = $i;
      $params->{mode} = 3;
      return TRUE;
    }
  }
  # unknown atom symbol
  return FALSE;
}

# end of number
sub endnumber
# returns TRUE if OK or FALSE if error
{
  my $params = shift;
  $params->{nmode} = FALSE;
  return FALSE unless ($params->{mode});
  if ($params->{mode} != 2)
  {
    if ($params->{mode} == 1)
    {
      return FALSE unless(endatom($params));
    }
    $params->{mode} = 0;
    $params->{atom}->[$params->{atoms}][1] *= $params->{numbers};
  }
#  elsif ($params->{mode} == 2)
  else
  {
    $params->{group}->[$params->{opengroup}->[$params->{opengroups} + 1]][0] *= $params->{numbers};
  }
  $params->{numbers} = "";
  return TRUE;
}

# space symbol
sub space
# returns TRUE if OK or FALSE if error
{
  my $params = shift;
  if ($params->{nmode})
  {
    return endnumber($params);
  }
  elsif ($params->{mode} == 1)
  {
    return endatom($params);
  }
  return TRUE;
}

# start new atom
sub newatom
# returns TRUE if OK or FALSE if error
{
  my $char = shift;
  my $params = shift;
  if ($params->{nmode})
  {
    return FALSE unless(endnumber($params));
  }
  elsif ($params->{mode} == 1)
  {
    return FALSE unless(endatom($params));
  }
  $params->{mode} = 1;
  ++$params->{atoms};
  $params->{atom}->[$params->{atoms}][0] = $char;
  $params->{atom}->[$params->{atoms}][1] = 1;
  $params->{atom}->[$params->{atoms}][2] = $params->{opengroup}->[$params->{opengroups}];
  return TRUE;
}

# reading atom symbol
sub atom
# returns TRUE if OK or FALSE if error
{
  my $char = shift;
  my $params = shift;
  if ($params->{mode} == 1)
  {
    unless ($params->{nmode})
    {
      my $number = $params->{uniqatoms};
      $params->{atom}->[$params->{atoms}][0] = $params->{atom}->[$params->{atoms}][0] . $char;
      return TRUE if (endatom($params));
      chop($params->{atom}->[$params->{atoms}][0]);
      --$params->{uniqatoms} if ($number > $params->{uniqatoms});
      return FALSE unless (endatom($params));
    }
  }
  return newatom(uc($char), $params);
}

# start new group
sub newgroup
# returns TRUE if OK or FALSE if error
{
  my $params = shift;
  if ($params->{nmode})
  {
    return FALSE unless(endnumber($params));
  }
  elsif ($params->{mode} == 1)
  {
    return FALSE unless(endatom($params));
  }
  $params->{mode} = 0;
  ++$params->{groups};
  $params->{group}->[$params->{groups}][0] = 1;
  $params->{group}->[$params->{groups}][1] = $params->{opengroup}->[$params->{opengroups}];
  ++$params->{opengroups};
  $params->{opengroup}->[$params->{opengroups}] = $params->{groups};
  return TRUE;
}

# end of group
sub endgroup
# returns TRUE if OK or FALSE if error
{
  my $params = shift;
  if ($params->{nmode})
  {
    return FALSE unless(endnumber($params));
  }
  elsif ($params->{mode} == 1)
  {
    return FALSE unless(endatom($params));
  }
  $params->{mode} = 2;
  --$params->{opengroups};
  return TRUE;
}

sub count
{
  my $gnum = shift;
  my $params = shift;
  my $count = 1;
  while ($gnum)
  {
    $count *= $params->{group}->[$gnum][0];
    $gnum = $params->{group}->[$gnum][1];
  }
  return $count;
}

# main subroutine
sub parse
# $_[0] = formula
# $_[1] = \%params
# returns TRUE if OK or FALSE if error
{
  my $formula = shift;
  my $params = shift;

  $formula =~ s/\s+/ /g;

  #check syntax
  { my $cnt; my $s;
    return FALSE unless (($formula =~ m/^[A-Za-z\d\.(){}\[\]\s]+$/) and
                ($formula !~ m/(?:\D+\.\D+|\d+\.\D+|\D+\.\d+|\d+\.\d+\.|\d+\s\d+|^\.|\.$)/) and
                ($formula =~ m/^
                  (?> (?{ $cnt = 0; $s = ""; })
                      (?:
                        (?> [^()[\]{}]+ )
                        |([\(\[\{]) (?{ ++$cnt; $s .= $1; })
                        |\) (?(?{ $cnt and (chop($s) eq "(") }) (?{ --$cnt; }) | (?!))
                        |] (?(?{ $cnt and (chop($s) eq "[") }) (?{ --$cnt; }) | (?!))
                        |} (?(?{ $cnt and (chop($s) eq "\{") }) (?{ --$cnt; }) | (?!))
                      )*
                  )
                  (?(?{ $cnt }) (?!))
                $/x));
  }

  my $error = FALSE;

  # char to char analysis
  foreach my $char (split //, $formula)
  {
    (newatom($char, $params)  ? next : ($error = TRUE, last)) if ($char =~ m/[A-Z]/);
    (atom($char, $params) ? next : ($error = TRUE, last)) if ($char =~ m/[a-z]/);
    (newgroup($params) ? next : ($error = TRUE, last)) if ($char =~ m/[(\{\[]/);
    (endgroup($params) ? next : ($error = TRUE, last)) if ($char =~ m/[\]\})]/);
    (number($char, $params), next) if ($char =~ m/[\d\.]/);
    ($error = TRUE, last) unless(space($params));
  }
  if (!($error) and ($params->{nmode}))
  {
    $error = !endnumber($params);
  }
  elsif (!($error) and ($params->{mode} == 1))
  {
    $error = !endatom($params);
  }
  return FALSE if ($error);

  for (my $i = 0; $i <= $params->{atoms}; ++$i)
  {
    for (my $j = 0; $j <= $params->{uniqatoms}; ++$j)
    {
      if ($params->{atom}->[$i][0] eq $params->{uniqatom}->[$j][0])
      {
        my $gnum = $params->{atom}->[$i][2];
        my $cnt = count($gnum, $params);
        $params->{uniqatom}->[$j][1] += $params->{atom}->[$i][1] * $cnt;
      }
    }

  }
  return TRUE;
}

sub round
{
  return(int($_[0] + 0.5));
}

# format value
sub prep
# input & output: $_[0] = value, $_[1] = absolute error
{
  unless ($_[1])
  {
    $_[0] = round($_[0]);
  }
  else
  {
    my ($head, $tail) = split /\./, sprintf("%f", $_[1]);
    if ($head)
    {
      if ($head == 1)
      {
        $_[1] = sprintf("%.1f", $_[1]);
        $_[0] = sprintf("%.1f", $_[0]);
      }
      else
      {
        $_[1] = round($_[1]);
        $_[0] = round($_[0]);
      }
    }
    else
    {
      $tail =~ m/[1-9]/g;
      my $pos = pos($tail);
      ++$pos if (substr($tail, $pos - 1, 1) == 1);
      my $format = "%." . $pos . "f";
      $_[1] = sprintf($format, $_[1]);
      $_[0] = sprintf($format, $_[0]);
    }
  }
}

# calculate molar mass and mass error
sub calcmass
{
  my $self = shift;

  # \@atoms
  my $atoms = shift;

  # Temporary values needed for calculations
  my %params = (
                uniqatoms => -1,
                uniqatom => [],
                atoms => -1,
                atom => [],
                groups => 0,
                group => [[1, 0]],
                opengroups => 0,
                opengroup => [],
                mode => 0,
                nmode => FALSE,
                numbers => ""
               );
  # array for uniq atoms: $uniqatom[$uniqatoms][0] = atom symbol, [1] = quantity,
  # [2] = atom number at @list
  # array for atoms: $atom[$atoms][0] = atom symbol, [1] = quantity, [2] = group number
  # array for groups: $group[$groups][0] = quantity, [1] = parent group number
  # array for groups opened: $opengroup[$opengroups] = group number
  # mode: 0 - none, 1 - atom, 2 - end of group, 3 - end of atom symbol
  # nmode: FALSE - none, TRUE - number or decimal point

  my $formula = $self->{formula};

  # apply acronyms
  for (my $i = 0; $i <= $#acronyms; ++$i)
  {
    $formula =~ s/$acronyms[$i][0](?=\d+)/($acronyms[$i][1])/gi;
    $formula =~ s/$acronyms[$i][0]/$acronyms[$i][1]/gi;
  }

  # change commas to points
  $formula =~ s/,/\./g;

  # coefficient rearrangement
  $formula =~ s/(?:^|\s*[*·]\s*)(\d+\.\d+|\d+)([^*·]+)/($2)$1/g;
  $formula =~ s/\s*[*·]\s*//g;

  my $error = 0;
  my $mass = 0;
  my $electrons = 0;

  # parse formula string
  if (parse($formula, \%params) and ($params{uniqatoms} >= 0))
  {
    for (my $i = 0; $i <= $params{uniqatoms}; ++$i)
    {
      # calculate molar mass
      $mass += $list[$params{uniqatom}->[$i][2]][2] * $params{uniqatom}->[$i][1];
      $electrons += $list[$params{uniqatom}->[$i][2]][0] * $params{uniqatom}->[$i][1];

      # check necessity of error calculation
      if (($error < 0) or ($list[$params{uniqatom}->[$i][2]][3]) eq "*")
      {
        $error = -1;
      }
      else
      {
        # calculate mass error
        $error += $list[$params{uniqatom}->[$i][2]][3] * $params{uniqatom}->[$i][1];
      }
    }
    $error = 0 if ($error < 0);
  }
  else
  {
    #return error if formula string is not valid
    return FALSE;
  }

  # format mass value
  prep($mass, $error);

  # return mass and error values
  $self->{mass} = $mass;
  $self->{error} = $error;

  # return number of electrons
  $self->{electrons} = $electrons;

  # return @atoms
  @{$atoms} = @{$params{uniqatom}};

  return  TRUE;
}

# calculate concentration of each element
sub calccon
# $_[1] = \@atoms
# input: $atoms[$i][0] = atom symbol, [1] = quantity, [2] = atom number at @list
# outpup: $atoms[$i][0] = atom symbol, [1] = quantity, [2] = concentration, [3] = concentration error
{
  my $self = shift;
  my $uniqatom = shift;
  my $uniqatoms = $#{$uniqatom};

  # relative error
  my $rerror = $self->{error} / $self->{mass};

  # atoms list
  my @atom = ();

  for (my $i = 0; $i <= $uniqatoms; ++$i)
  {
    # atom symbol
    $atom[$i][0] = $uniqatom->[$i][0];

    # quantity
    $atom[$i][1] = $uniqatom->[$i][1];

    # concentration
    $atom[$i][2] = $list[$uniqatom->[$i][2]][2] * $uniqatom->[$i][1] * 100 / $self->{mass};

    # check necessity of error calculation
#    unless ($_[1] < 0)
    if ($rerror)
    {
      # absolute error
      $atom[$i][3] = $atom[$i][2] * ($rerror + $list[$uniqatom->[$i][2]][3] /
                        $list[$uniqatom->[$i][2]][2]);

      # format concentration value
      prep($atom[$i][2], $atom[$i][3]);
    }
    else
    {
      # no error
      $atom[$i][3] = 0;

      # format concentration value
      $atom[$i][2] = sprintf("%.1f", $atom[$i][2]);
    }
  }
  # return @atom
  @{$uniqatom} = @atom;
}

# main function
# called from mmcalc and gmmcalc scripts
# calculate molar mass, concentrations and errors
# returns FALSE if molar mass was not calculated
# or TRUE if done
sub calculate
{
  my $self = shift;
  $self->reset() if ($self->{formula});
  $self->{formula} = shift;
  my @atoms = ();

  # calculate molar mass and mass error
  if ($self->calcmass(\@atoms))
  {
    # calculate concentrations and errors
    $self->calccon(\@atoms);
    # $atoms[$i][0] = atom symbol, [1] = quantity,
    # [2] = concentration, [3] = absolute error
    # sort atoms list
    @{$self->{atoms}} = sort {$a->[0] cmp $b->[0]} @atoms;

    $self->{content} = $self->get_formula();

    return TRUE;
  }
  else
  {
    # return error if molar mass was not calculated
    return FALSE;
  }
}


# Operations with atoms and acronyms lists:

sub check_atom
# returns TRUE if atom exist
# $_[0] = atom symbol
{
  for (my $i = 0; $i <= $#list; ++$i)
  {
    ($_[0] = $list[$i][1], return TRUE) if ($_[0] =~ m/^$list[$i][1]$/i);
  }
  return FALSE;
}

sub check_acronym
# returns TRUE if OK or FALSE if error
# IN:
# $_[0] = \@list
# $_[1] = acronym label
# $_[2] = acronym formula
# $_[3] = comments
# OUT:
# $_[4] = error message if error
{
  # check acronym label
  ($_[4] = "Invalid acronym: \"$_[1]\"!", return FALSE) unless ($_[1] =~ m/^[A-Za-z]{2,10}$/);

  # check comments
  ($_[4] = "Invalid comments: \"$_[3]\"!", return FALSE)
      unless ($_[3] =~ m/^[A-Za-z\d()\[\].,\-\+ ]{1,100}$/);

  # check for collision of acronym and element symbol
  ($_[4] = "Invalid acronym: collision with atom $_[1]!", return FALSE)
      if ((length($_[1]) < 3) and check_atom($_[1]));

  # check acromym duplication
  for (my $j = 0; $j <= $#{$_[0]}; ++$j)
  {
    ($_[4] = "Invalid acronym: \"$_[1]\" already exist!", return FALSE)
        if ($_[1] =~ m/^${$_[0]}[$j][0]$/i);
  }

  # check formula validity
  ($_[4] = "Invalid acronym: uncorrect formula \"$_[2]\"!", return FALSE) unless (parse($_[2]));

  return TRUE;
}

sub set_acronyms
# $_[0] = \@list
{
  my @list = ();
  my $i = 0;

  foreach my $item (@{$_[0]})
  {
    # check validity of acronym
    my $error_message;
    next unless (check_acronym(\@list, $item->[0], $item->[1], $item->[2], $error_message));
    # add acronym to acronyms list
    $list[$i][0] = $item->[0];  # acronym
    $list[$i][1] = $item->[1];  # formula
    $list[$i][2] = $item->[2];  # name or comments
    ++$i;
  }

  # sort list
  @list = sort { my $flag;
      ($flag = (length(${$b}[0]) <=> length(${$a}[0]))), ($flag ? $flag : (lc(${$a}[0]) cmp lc(${$b}[0])))
    } @list;

  # set new acronyms list if it is different from current list
  my $flag = FALSE;
  if (($#acronyms >= 0) and ($#acronyms == $#list))
  {
    $flag = TRUE;
    for (my $i = 0; $i <= $#acronyms; ++$i)
    {
      ($flag = FALSE, last) if (($list[$i][0] ne $acronyms[$i][0]) and
                                ($list[$i][1] ne $acronyms[$i][1]) and
                                ($list[$i][2] ne $acronyms[$i][2]));
    }
  }
  @acronyms = @list unless ($flag);
}

sub get_acronyms
{
  @{$_[0]} = @acronyms;
}

sub write_acronyms
# if OK: returns TRUE, $_[0] = file path
# if nothing to save: returns FALSE, $_[0] = ""
# if failed: returns FALSE, $_[0] = path
{
  $_[0] = "";
  umask(0022);

  # nothing to save
  return FALSE unless ($#acronyms >= 0);

  unless ($acronyms_file and -f($acronyms_file) and -T _ and -w _)
  {
    my $path = data_home('mmcalc');

    # resolv symlink
    $path = realpath($path) if (-e($path) and -l($path));

    # remove broken symlink
    (unlink($path) or die "$!\n") if (!(-e($path)) and -l($path));

    $_[0] = $path;

    # not a directory
    return FALSE if (-e($path) and !(-d _));

    # not exist directory
    unless ($path and -e($path) and -d _)
    {
      return FALSE unless (-e(data_home()) and -d _ and -w _);
      mkpath($path);
    }

    # write protected directory
    return FALSE unless ($path and -w($path));

    $acronyms_file = catfile($path, 'acronyms.dat');
  }

  # write protected file
  return FALSE if (-e($acronyms_file) and !(-w _));

  open(ACRONYMS, ">", $acronyms_file) or die "open file error: $!\n";
  print ACRONYMS "# acronyms list\n#acronym\tformula\t\tname or comments\n";
  foreach (@acronyms)
  {
    print ACRONYMS "${$_}[0]\t\t${$_}[1]\t\t${$_}[2]\n";
  }
  close ACRONYMS or die "$!\n";
  $_[0] = $acronyms_file;
  return TRUE;
}


# Operations with file names:

sub search_file
# IN:
# $_[0] = file name
# $_[1] = search dirs
# OUT: full path to file or undef
{
  my $file;
  my @search_dirs = @{$_[1]};
  foreach my $dir (@search_dirs)
  {
    my $tmp = catfile($dir, $_[0]);
    ($file = $tmp, last) if (-e($tmp) and -f _ and -T _ and -r _);
  }
  return $file;
}

sub get_file
# IN:  $_[0] = file name
# OUT: full path to file or undef
{
  # look for the file in $XDG_DATA_HOME/mmcalc/
  my $file = data_files('mmcalc', $_[0]);

  #returns $file if defined and can be read
  return $file if (defined($file) and -f($file) and -T _);

  # look for the file in $PREFIX/share/mmcalc, script dir or current dir
  my @searchdirs = (catfile($prefix, 'share', 'mmcalc'), $script_dir, $current_dir);
  search_file($_[0], \@searchdirs);

  # returns full path to file or undef
}


# Initialisation:

sub init
{
  # atoms list
  @list = ();

  # get full path to file of atoms list
  $atoms_file = get_file("atoms.dat");
  die "File atoms.dat not found.\n" unless (defined($atoms_file));

  # read atoms list from file
  open(ATOMS, $atoms_file) or die "$!: $atoms_file\n";
  my $i = 0;
  while(<ATOMS>)
  {
    chomp;
    s/^\s+//;
    s/\s+$//;

    # check line validity
    # valid line examples:
    # 1       H       Hydrogen        1.00794         0.00007
    # 43      Tc      Technetium      98              *
    if (m/^\d{1,3}\s+[A-Z][a-z]?\s+[A-Za-z]+\s+\d+\.?\d+\s+(\d+\.\d+|\*)$/)
    {
      my @args = (split);

      # check atoms duplication
      next if (check_atom($args[1]));

      # add atom to atoms list
      $list[$i][0] = $args[0];  # atom number
      $list[$i][1] = $args[1];  # atom symbol
      $list[$i][2] = $args[3];  # atom mass
      $list[$i][3] = $args[4];  # mass absolute error
      ++$i;
    }
  }
  close ATOMS or die "$!";

  # Number of atoms less than 112.
  die "Number of atoms less than 112.\n" if ($i < 112);

  # get full file name for acronyms list
  $acronyms_file = get_file("acronyms.dat");
  if (defined($acronyms_file))
  {
    # new acronyms list
    my @alist = ();

    # read acronyms list from file
    open(ACRONYMS, $acronyms_file) or die "$!: $acronyms_file\n";
    $i = 0;
    while(<ACRONYMS>)
    {
      chomp;
      s/#.*$//;
      s/^\s+//;
      s/\s+$//;

      # check line validity
      # valid line examples:
      # dabco    C6H12N2       diazabicyclo[2.2.2]octane
      # def      (C2H5)2NCHO   diethylformamode
      if (m/^[A-Za-z]{2,10}\t+[A-Z(][A-Za-z\d.()]{1,100}\t+[A-Za-z\d()\[\].,\-\+ ]{1,100}$/)
      {
        my @args = (split /\t+/);

        $alist[$i][0] = $args[0];  # acronym
        $alist[$i][1] = $args[1];  # formula
        $alist[$i][2] = $args[2];  # name or comment
        ++$i;
      }
    }
    close ACRONYMS or die "$!";

    # set acronyms list
    @acronyms = ();
    set_acronyms(\@alist);
  }
  else
  {
    warn "WARNING: file acronyms.dat not found.\n";
  }
  return TRUE;
}

1;
